import functools
import weakref
from typing import Any, Callable, Optional, Type, TypeVar, Union, cast
from typing_extensions import Protocol
from typing import Callable, TypeVar, Generic
'''
Typing this library correctly is impossible right now.

The main issue is that we can not create types for callable objects with a partial signature, like "Callable[[SELF, ...], OUT]" for a normal method.

https://github.com/python/mypy/issues/5876

'''

_T = TypeVar('_T')


class _NormalMethod(Protocol[_T]):
    __name__: str
    __call__: Callable[..., _T]


_ClassMethod = _NormalMethod
_StaticMethod = _NormalMethod


class _Property(Protocol[_T]):
    __get__: Callable[..., _T]


class cached_property(property, _Property[_T]):
    '''
    decorator for converting a method into a cached property
    See https://stackoverflow.com/a/4037979/3671939
    This uses a modification:
    1. inherit from property, which disables setattr(instance, name, value)
        as it raises AttributeError: Can't set attribute
    2. use instance.__dict__[name] = value to fix
    '''

    def __init__(self, method: Callable[..., _T]):
        self._method = method

    def __get__(self, instance, _) -> _T:
        name = self._method.__name__
        value = self._method(instance)
        instance.__dict__[name] = value
        return value


def set_method(cls):
    '''decorator for adding or replacing a method of a given class'''
    return _set_method(cls, None)


def _set_method(cls, proxy: Callable = None):
    '''decorator for adding or replacing a method of a given class'''

    def decorator(method: _NormalMethod):
        assert hasattr(method, '__call__'), f'Not callable method: {method}'
        name: str = method.__name__
        method = proxy(method) if proxy else method
        setattr(cls, name, method)

    return decorator


def cached_method(maxsize=128, typed=False):
    '''decorator for converting a method into an lru cached method'''

    # https://stackoverflow.com/a/33672499/3671939
    def decorator(func: _NormalMethod[_T]) -> _NormalMethod[_T]:

        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            weak_self = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(maxsize=maxsize, typed=typed)
            def _cached_method(*args, **kwargs):
                self = weak_self()
                assert self
                return func(self, *args, **kwargs)

            setattr(self, func.__name__, _cached_method)
            return _cached_method(*args, **kwargs)

        return wrapped_func

    return decorator


def set_cached_method(cls, maxsize=128, typed=False):
    '''decorator for adding or replacing a cached_method of a given class'''
    return _set_method(cls, cached_method(maxsize, typed))


def set_classmethod(cls):
    '''decorator for adding or replacing a classmethod of a given class'''
    return _set_method(cls, classmethod)


def set_staticmethod(cls):
    '''decorator for adding or replacing a staticmethod of a given class'''
    return _set_method(cls, staticmethod)


def set_property(cls):
    '''decorator for adding or replacing a property of a given class'''
    return _set_method(cls, property)


def set_cached_property(cls):
    '''decorator for adding or replacing a cached_property of a given class'''
    return _set_method(cls, cached_property)


set_cachedproperty = set_cached_property  # Previous versions

F = TypeVar('F', bound=Callable[..., Any])


class copy_signature(Generic[F]):

    def __init__(self, target: F) -> None:
        ...

    def __call__(self, wrapped: Callable[..., Any]) -> F:
        return cast(F, wrapped)


from typing import Callable, TypeVar, Generic
import functools, types

F = TypeVar('F', bound=Callable[..., Any])


def copy_function(f: F, name: str = None) -> F:
    """Based on http://stackoverflow.com/a/6528148/190597"""
    g = types.FunctionType(
        f.__code__,
        f.__globals__,
        name=name or f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return cast(F, g)


class implemented_at(Generic[F]):
    '''
    Decorator that ignores the wrapped function and uses the
    parameter function instead.
    
    Useful for implementing methods in separate files.
        Docs are copied so that any editor can provide help(method).
        Types signatures are copied so that type checkers work seamlessly.
    
    For usage details see help(implemented_at.usage)
    '''

    def __init__(self, target: F) -> None:
        self.target = target

    def __call__(self, wrapped: Callable[..., Any]) -> F:
        return copy_function(self.target, wrapped.__name__)

    @staticmethod
    def usage():
        '''
        # --- main.py ---
        (from .) import other_file

        class Hello:
            @implemented_at(other_file.hello)
            def hello(self):
                ...

        # --- other_file.py ---
        from __future__ import annotations
        from typing import TYPE_CHECKING, Type
        if TYPE_CHECKING:
            from (.)main import Hello

        def hello(obj: Hello, name:str):
            'I just say "hello {name}" for some given name'
            print(f'hello {name}')

        # --- interactive console ---
        h = Hello()
        help(h.hello)
        h.hello('world')
        '''


if False:

    class Test:
        zero = 0

        @cached_property
        def prop(self):
            print('x')
            return 1 + 1

    @set_method(Test)
    def f(self: Test):
        print(self.zero)

    t = Test()
    x = t.prop


import sys
import importlib
from types import ModuleType

# https://stackoverflow.com/q/42870428


def deep_reload(m: ModuleType):
    name = m.__name__  # get the name that is used in sys.modules
    name_ext = name + '.'  # support finding sub modules or packages
    del m

    def compare(loaded: str):
        return (loaded == name) or loaded.startswith(name_ext)

    all_mods = tuple(
        sys.modules)  # prevent changing iterable while iterating over it
    sub_mods = filter(compare, all_mods)
    for pkg in sub_mods:
        del sys.modules[
            pkg]  # remove sub modules and packages from import cache
    return importlib.import_module(name)


from __future__ import annotations
from typing import List, Optional, Tuple
import abc
from .types import (Params, Data, unzip)


class _Str(abc.ABC):

    @abc.abstractmethod
    def _str(self, params: Params) -> str:
        ...


class QueryBody(_Str):

    _where: Optional[Where] = None
    _order_by: Tuple[str, ...] = ()
    _limit: Optional[int] = None

    def _str(self, params: List[Data]):
        body = []
        if self._where is not None:
            where_str = self._where._str(params)
            body.append(f'WHERE {where_str}')
        if self._order_by:
            by = ', '.join(self._order_by)
            body.append(f'ORDER BY {by}')
        if self._limit is not None:
            body.append('LIMIT ?')
            params.append(int(self._limit))
        return ' '.join(body)


class Where(_Str):

    @staticmethod
    def assert_initialized(where: Optional[Where]):
        assert where is not None, (
            'To prevent an accident, you must specify'\
            ' .where(..) for delete/update queries.'
            ' If you really want everywhere use '
            ' .everywhere()'
        )


class Everywhere(Where):

    def _str(self, params: Params):
        return ''


class WhereSql(Where):

    def __init__(self, query: str, *params: Data):
        self.query = query
        self.params = params

    def _str(self, params: Params):
        params.extend(self.params)
        return self.query


class WhereEqual(WhereSql):
    '''
    Particular case of Where:
        key1 = value1 AND key2 = value2 AND ...
    Implemented separately for speed because it is very common 
    '''

    def __init__(self, **kwargs: Data):
        keys, values = unzip(kwargs)
        query = ' AND '.join(f'{k} = ?' for k in keys)
        super().__init__(query, *values)
        # save also kwargs (only for update query):
        self.kwargs = kwargs


class WhereAnd(Where):

    def __init__(
        self,
        first: Optional[Where],
        second: Where,
    ):
        assert first, 'You can not use .and_where_sql(..) without .where(..) before'
        self.first = first
        self.second = second

    def _str(self, params: Params):
        first = self.first._str(params)
        second = self.second._str(params)
        return f'({first}) AND ({second})'



from __future__ import annotations
from pathlib import Path
import sqlite3
#from sqlite3.dbapi2 import ProgrammingError
from typing import List, Any, Tuple, Union, cast
from pathlib import Path
from .types import (
    Data,
    Params,
)

FilePath = Union[str, Path]


def custom_repr(self, *keys):
    name = self.__class__.__name__
    comma_kwargs = ', '.join(f'{k}={repr(getattr(self, k))}' for k in keys)
    return f'{name}({comma_kwargs})'


class SqliteDB:
    '''
    '''

    def __init__(self, file: FilePath):
        self.file = file

    def __repr__(self):
        return custom_repr(self, 'file')

    def new_connection(self):
        return sqlite3.connect(self.file, check_same_thread=False)

    def _execute(self, query: str, params: Params = None) -> sqlite3.Cursor:
        assert not isinstance(params, str), f'Did you mean params=[{params}]?'
        try:
            with self.new_connection() as con:
                return con.execute(query, params or [])
        except sqlite3.Error as e:
            e.args = (*e.args, query, params)
            raise e

    def execute(self, query: str, params: Params = None) -> List[List[Data]]:
        return [*self._execute(query, params or [])]

    def execute_column(self, query: str, params: Params = None) -> List[Data]:
        rows = self.execute(query, params)
        return [first for first, *_ in rows]

    def get_table(self, table_name: str):
        from .table import SqliteTable
        return SqliteTable(self.file, table_name)

    def table_names(self) -> List[str]:
        table_names = self.execute_column("""
            SELECT name FROM sqlite_master 
            WHERE type = 'table' 
            AND name NOT LIKE 'sqlite_%'
            ORDER BY 1;
        """)
        table_names = cast(List[str], table_names)
        return table_names

    def index_names(self):
        return self.execute_column("""
            SELECT name FROM sqlite_master 
            WHERE type = 'index' 
            ORDER BY 1;
        """)



from __future__ import annotations
from typing import Any, Generic, List, Optional, Tuple, Type, TypeVar, TypedDict, cast, overload
from contextlib import contextmanager
import json, time, random, logging
from .table import SqliteTable, FilePath


class StoreRecord(TypedDict):
    key: str
    value: Optional[str]
    lock_token: float
    locked_until: float


def create_deadline(timeout: Optional[float]):
    if timeout is None:
        return float('inf')
    return time.time() + timeout


class TimeoutError(Exception):
    message = 'Waiting for access token timed out'


class TokenError(Exception):
    message = 'Invalid or expired token'


Value = TypeVar('Value')
Data = TypeVar('Data')


class SqliteStore(Generic[Value], SqliteTable):

    def __init__(self, file: FilePath, name: str):
        self.table.__init__(file, name)
        self.db.execute(f'''
        CREATE TABLE IF NOT EXISTS {self.name}(
            key text NOT NULL PRIMARY KEY,
            value text,
            lock_token double NOT NULL,
            locked_until double NOT NULL
        )
        ''')
        the_columns = {'key', 'value', 'lock_token', 'locked_until'}
        assert set(self.columns()) == the_columns, self.columns()
        return

    @property
    def table(self):
        return cast(SqliteTable, super())

    def __getitem__(self, key: str):
        d = self.table.where(key=key).get_dict(
            'value',
            type=StoreRecord,
        )
        d = cast(Optional[StoreRecord], d)
        if not d or not d['value']:
            raise KeyError(key)
        value = json.loads(d['value'])
        return cast(Value, value)

    @overload
    def get(self, key: str) -> Optional[Value]:
        ...

    @overload
    def get(self, key: str, default: Value = None) -> Optional[Value]:
        ...

    def get(self, key: str, default=None):
        try:
            value = self[key]
        except KeyError:
            value = default
        return value

    def getT(self, key: str, default=None, type: Type = Any):
        try:
            value = self[key]
        except KeyError:
            value = default
        return cast(type, value)

    def values(self):
        col = self.table.where_sql(
            'value<>?',
            None,
        ).column('value', type=str)
        values = [json.loads(s) for s in col]
        return cast(List[Value], values)

    def keys(self):
        return self.table.column('value', type=str)

    def items(self):
        encoded = self.table.rows(
            'key',
            'value',
            type=Tuple[str, Optional[str]],
        )
        items = [(k, json.loads(v or 'null')) for k, v in encoded]
        items = cast(List[Tuple[str, Value]], items)
        return items

    def __set_assuming_token(self, key: str, value: Value):
        return self.table.where(key=key).update(value=json.dumps(value))

    def __del_assuming_token(self, key: str):
        return self.table.where(key=key).delete()

    def _current_lock(self, key: str):
        return self.table.where(key=key).get_dict(
            'lock_token',
            'locked_until',
            type=StoreRecord,
        )

    def set(self, key: str, value: Value, token: float):
        '''
        Requires a token provided by the exclusive access context
        manager self.wait_token() or self.ask_token().
        '''
        with self.assert_token(key, token=token):
            self.__set_assuming_token(key, value)
        return

    def delete(self, key: str, token: float):
        '''
        Requires a token provided by the exclusive access context
        manager self.wait_token() or self.ask_token().
        '''
        with self.assert_token(key, token=token):
            self.__del_assuming_token(key)
        return

    def ask_del(self, key: str):
        with self.ask_token(key) as token:
            if not token:
                return False
            self.__del_assuming_token(key)
        return False

    def ask_set(self, key: str, value: Value):
        with self.ask_token(key) as token:
            if not token:
                return False
            self.__set_assuming_token(key, value)
        return False

    def wait_del(self, key: str, request_every: float = 0.02,
                 timeout: Optional[float] = 3):
        wait = self.wait_token(
            key,
            request_every=request_every,
            timeout=timeout,
        )
        with wait:
            self.__del_assuming_token(key)

    def wait_set(
        self,
        key: str,
        value: Value,
        request_every: float = 0.02,
        timeout: Optional[float] = 3,
    ):
        wait = self.wait_token(
            key,
            request_every=request_every,
            timeout=timeout,
        )
        with wait as token:
            self.set(key, value, token)

    @contextmanager
    def ask_token(
        self,
        *keys: str,
        max_duration: float = 0.5,
    ):
        '''
        with table.ask_token('some_key') as token:
            if not token:
                # The resource is locked by other
            else:
                # The resource is mine
                table.set('some_key', some_value, token)
        '''
        assert keys, 'You must specify keys to be locked explicitely'
        token = random.random()
        try:
            gained_access = all([
                self._ask_access(key, token=token, max_duration=max_duration)
                for key in keys
            ])
            yield token if gained_access else None
        finally:
            for key in keys:
                self._unlock(key, token, max_duration)
        return

    @contextmanager
    def assert_token(self, *keys: str, token: float):
        assert keys, 'You must specify keys to be locked explicitely'
        max_duration = 0.5
        try:
            for key in keys:
                if not self._ask_access(
                        key,
                        token=token,
                        max_duration=max_duration,
                ):
                    raise TokenError(key, token)
            yield
        finally:
            for key in keys:
                self._unlock(key, token, max_duration)
        return

    @contextmanager
    def wait_token(
        self,
        *keys: str,
        max_duration: float = 0.5,
        request_every: float = 0.02,
        timeout: Optional[float] = 3,
    ):
        '''
        # Wait at most {timeout} seconds to get exclusive access,
        # requesting every {request_every} seconds
        with table.wait_token('some_key') as token:
            current = table['some_key']
            ...
            # No one else can set some_key
            # during next {max_duration} seconds
            ...
            current = table.set('some_key', some_value, token)
        '''
        assert keys, 'You must specify keys to be locked explicitely'
        token = random.random()
        try:
            for key in keys:
                deadline = create_deadline(timeout)
                while not self._ask_access(key, token, max_duration):
                    if time.time() > deadline:
                        raise TimeoutError(timeout)
                    time.sleep(request_every)
            yield token
        finally:
            for key in keys:
                self._unlock(key, token, max_duration)
        return

    def _ask_access(self, key: str, token: float, max_duration: float):
        now = time.time()
        until = now + max_duration
        # Compete with other processes for an exclusive update
        did = self.table.where(key=key).and_where_sql(
            """
            lock_token<0 OR
            lock_token=? OR
            locked_until<?
            """,
            token,
            now,
        ).update_or_ignore(
            lock_token=token,
            locked_until=until,
        )
        if did:  # Race winner
            return True
        # Maybe the key was not even present:
        return self.insert_or_ignore(
            key=key,
            lock_token=token,
            locked_until=until,
        )

    def _unlock(self, key: str, token: float, max_duration: float):
        d = self._current_lock(key)
        if not d:
            # Entry was deleted (and unlocked)
            return
        remaining = d['locked_until'] - time.time()
        if remaining < 0:
            ms = round(-remaining * 1000)
            logging.warning(
                f'Locked {repr(key)} during {ms}ms more than max_duration={max_duration}s'
            )
        if d['lock_token'] == token:
            self.table.where(key=key).update(lock_token=-1)
        return



from __future__ import annotations
from typing import Callable, Dict, Iterator, List, Any, Optional, Type, TypeVar, cast
from functools import wraps
from random import shuffle
from . import query_body as QB
from .types import (DataRow, Data, Record, unzip)
from .database import SqliteDB

F = TypeVar('F', bound=Callable)


def query_method(wrapped: F) -> F:
    '''
    Decorator for using TableQuery methods
    from classes that inherit from it.
    When called from a parent, self is converted
    into a new TableQuery with a new empty _body.
    '''

    @wraps(wrapped)
    def wrapper(self: TableQuery, *args, **kwargs):
        if not hasattr(self, '_body'):
            self = TableQuery.new(self)
        return wrapped(self, *args, **kwargs)

    return cast(F, wrapper)


class TableQuery:

    # Abstract attributes (defined by parents):
    db: SqliteDB
    name: str  # table name

    # Non-abstract:
    _body: QB.QueryBody

    @classmethod
    def new(cls, parent: TableQuery):
        query = cls()
        query._body = QB.QueryBody()
        query.db = parent.db
        query.name = parent.name
        return query

    # "Where" related methods

    @query_method
    def where(self, **key_values):
        self._body._where = QB.WhereEqual(**key_values)
        return self

    @query_method
    def where_sql(self, query: str, *params: Any):
        self._body._where = QB.WhereSql(query, *params)
        return self

    @query_method
    def order_by(self, *columns_asc_desc: str):
        self._body._order_by = columns_asc_desc
        return self

    @query_method
    def limit(self, limit: Optional[int]):
        self._body._limit = limit
        return self

    @query_method
    def everywhere(self):
        self._body._where = QB.Everywhere()
        return self

    @query_method
    def and_where(self, **key_values):
        self._body._where = QB.WhereAnd(
            self._body._where,
            QB.WhereEqual(**key_values),
        )
        return self

    @query_method
    def and_where_sql(self, query: str, *params: Any):
        self._body._where = QB.WhereAnd(
            self._body._where,
            QB.WhereSql(query, *params),
        )
        return self

    # Select method and mutation methods

    @query_method
    def select(self, *columns: str):
        what = ', '.join(columns) or '*'
        params = []
        where = self._body._str(params)
        query = f'SELECT {what} FROM {self.name} {where}'
        return self.db._execute(query, params)

    @query_method
    def delete(self):
        QB.Where.assert_initialized(self._body._where)
        params = []
        where = self._body._str(params)
        query = f'DELETE FROM {self.name} {where}'
        return self.db._execute(query, params)

    @query_method
    def insert(self, **record: Data):
        #inst = 'INSERT OR IGNORE' if ignore else 'INSERT'
        where = self._body._str([]).strip()
        assert not where, f'Unexpected statements for INSERT: {where}'
        table = self.name
        keys, values = unzip(record)
        columns = ', '.join(keys)
        marks = ', '.join('?' for _ in keys)
        query = f'INSERT INTO {table} ({columns}) VALUES ({marks})'
        return self.db._execute(query, values)

    @query_method
    def insert_or_ignore(self, **record: Data):
        where = self._body._str([]).strip()
        assert not where, f'Unexpected statements for INSERT: {where}'
        table = self.name
        keys, values = unzip(record)
        columns = ', '.join(keys)
        marks = ', '.join('?' for _ in keys)
        query = f'INSERT OR IGNORE INTO {table} ({columns}) VALUES ({marks})'
        cursor = self.db._execute(query, values)
        return cursor.rowcount

    @query_method
    def update_or_ignore(self, **partial_record: Data):
        QB.Where.assert_initialized(self._body._where)
        keys, params = unzip(partial_record)
        what = ', '.join(f'{c} = ?' for c in keys)
        where = self._body._str(params)
        query = f'UPDATE {self.name} SET {what} {where}'
        cursor = self.db._execute(query, params)
        return cursor.rowcount

    @query_method
    def update(self, **partial_record: Data):
        if not self.update_or_ignore(**partial_record):
            assert isinstance(
                self._body._where, QB.WhereEqual
            ), '.where_sql(..).update(..) is unsupported. Use .where(..).update(..) instead'
            record = {
                **partial_record,
                **self._body._where.kwargs,
            }
            self._body._where = None
            self.insert(**record)
        return

    # high-level select methods

    @query_method
    def count(self):
        n = self.value('count(*)')
        return cast(int, n)

    # High-level methods with type argument

    @query_method
    def rows(self, *columns: str, type: Type = DataRow):
        rows = [*self.select(*columns)]
        return cast(List[type], rows)

    @query_method
    def dicts(self, *columns: str, type: Type = Record):
        cursor = self.select(*columns)
        names = [c for c, *_ in cursor.description]
        dicts = [dict(zip(names, row)) for row in cursor]
        return cast(List[type], dicts)

    @query_method
    def column(self, column: str, type: Type = Data):
        values = [first for first, *_ in self.rows(column)]
        return cast(List[type], values)

    @query_method
    def value(self, column: str, type: Type = Data):
        value = self.limit(1).column(column)[0]
        return cast(type, value)

    @query_method
    def dict(self, *columns: str, type: Type = Record):
        dict = self.limit(1).dicts(*columns)[0]
        return cast(type, dict)

    @query_method
    def get_value(self, column: str, type: Type = Data):
        col = self.limit(1).column(column, type=type)
        value = col[0] if col else None
        return cast(Optional[type], value)

    @query_method
    def get_dict(self, *columns: str, type: Type = Record):
        dicts = self.limit(1).dicts(*columns, type=type)
        record = dicts[0] if dicts else None
        return cast(Optional[type], record)

    @query_method
    def series(self, *columns: str, type: Type = Data):
        cursor = self.select(*columns)
        keys = [x[0] for x in cursor.description]
        out = {key: [] for key in keys}
        for row in cursor:
            for key, value in zip(keys, row):
                out[key].append(value)
        return cast(Dict[str, List[type]], out)

    # random selection methods

    @query_method
    def random_rows(self, n: int, *columns: str, type: Type = DataRow):
        rows, _ = self._select_random(n, *columns)
        return cast(List[type], rows)

    @query_method
    def random_dicts(self, n: int, *columns: str, type: Type = Record):
        rows, keys = self._select_random(n, *columns)
        dicts = [dict(zip(keys, row)) for row in rows]
        return cast(List[type], dicts)

    @query_method
    def random_values(self, n: int, column: str, type: Type = Data):
        rows = self.random_rows(n, column)
        values = [first for first, *_ in rows]
        return cast(List[type], values)

    @query_method
    def random_dict(self, *columns: str, type: Type = Record):
        dicts = self.random_dicts(1, *columns, type=type)
        return cast(type, dicts[0])

    @query_method
    def random_value(self, column: str, type: Type = Data):
        values = self.random_values(1, column, type=type)
        return cast(type, values[0])

    @query_method
    def _select_random(self, n: int, *columns: str):
        assert n > 0, 'Empty selection unsupported'
        N: int = self.count()
        assert N > 0, 'The table is empty'
        where = self._body._where
        column_names = []
        rows: List[List[Data]] = []
        while len(rows) < n:
            remaining = n - len(rows)
            # Filter many out randomly
            mod = 1 + N // (2 * remaining)
            print(mod)
            new_where = QB.WhereSql(f'random() % ? = 0', mod)
            if where is not None:
                new_where = QB.WhereAnd(where, new_where)
            self._body._order_by = ()
            self._body._where = new_where
            cursor = self.limit(remaining).select(*columns)
            column_names = [x[0] for x in cursor.description]
            rows.extend(cursor)
        shuffle(rows)
        return rows, column_names

    # high-level iter_select methods

    @query_method
    def iter_rows(self, *columns: str, type: Type = DataRow):
        total_approx = self.count()
        it = iter(self.select(*columns))
        it = cast(Iterator[type], it)
        return total_approx, it

    @query_method
    def iter_dicts(self, *columns: str, type: Type = Record):
        total_approx = self.count()
        cursor = self.select(*columns)
        names = [c for c, *_ in cursor.description]
        it = (dict(zip(names, row)) for row in cursor)
        it = cast(Iterator[type], it)
        return total_approx, it

    @query_method
    def iter_column(self, column: str, type: Type = Data):
        total_approx = self.count()
        it = iter(row[0] for row in self.select(column))
        it = cast(Iterator[type], it)
        return total_approx, it



from __future__ import annotations
from typing import Dict, List, Optional, cast
from random import shuffle
from .database import SqliteDB, FilePath, custom_repr
from .query import TableQuery


class SqliteTable(TableQuery):
    '''
    Sqlite table object.

    count(**kw) -> int
    columns() -> List[str]

    dicts(columns, **kw)      -> List[Record] : (Record=Dict[str, Data])
    column(column, **kw)      -> List[Data]
    series(columns, **kw)     -> Dict[str, List[Data]]
    rows(columns, **kw)       -> List[Tuple[Data]]
    
    dict(columns, **kw)       -> first Record (or Exception)
    get_dict(columns, **kw)   -> first Record | None

    value(column, **kw)       -> first value (or Exception)
    get_value(column, **kw)   -> first value | None
    '''

    def __init__(self, file: FilePath, table_name: str):
        self.file = file
        self.name = table_name
        self.db = SqliteDB(self.file)

    def __repr__(self):
        return custom_repr(self, 'file', 'table_name')

    def __len__(self):
        return self.count()

    def columns(self) -> List[str]:
        query_str = 'SELECT name FROM PRAGMA_TABLE_INFO(?)'
        column_names = self.db.execute_column(
            query_str,
            [self.name],
        )
        column_names = cast(List[str], column_names)
        return column_names


def test():
    import random
    table = SqliteTable('.test.db', 'rand_table')
    table.db.execute(f'''
        CREATE TABLE IF NOT EXISTS rand_table(
            key int NOT NULL PRIMARY KEY,
            age int NOT NULL,
            prob double NOT NULL
        )
    ''')

    print(len(table))
    for k in set(range(1000)).difference(set(table.column('key'))):
        table.insert(
            key=k,
            age=random.randint(0, 5),
            prob=round(random.random(), 3),
        )
    print(table.random_dicts(3))
    print(table.random_dicts(3))
    print(len(table))



'''
Small alternative to docopt and click for creating Command Line Interfaces (CLIs) out of functions:

 - Also composable as click (using CommandDict instead of MultiCommand)
 - Usage for commands in CommandDict's usage is shown by default
 - Command defaults are shown by default
 - Environment variables can be used optionally and are shown by default

Example:

from command_line import Command, Option, CommandDict, parse_port

def add_numbers(first:int, second:int):
    return first + second

def serve(port:int, host:str, dev:bool):
    print(port, host, dev)
    return

command_line = CommandDict(
    serve=Command(
        serve,
        Option('host', default='localhost'),
        Option('port', env='PORT', default='5000', parser=parse_port),
        Option('dev/prod', env='DEV', default='1'),
    ),
    add=Command(
        add_numbers,
        Option('first', positional=True, parser=int),
        Option('second', positional=True, parser=int),
    ),
)

if __name__ == '__main__':
    command_line()
'''
from __future__ import annotations
import os, sys
import abc
from typing import Any, Callable, Dict, Generic, List, Literal, Optional, Sequence, Tuple, Type, TypeVar, TypedDict, Union
import pprint


def parse_port(s: str):
    port = int(s)
    assert 1024 <= port <= 65535, (f'Port {port} out of range [1024, 65535]')
    return port


class Option:

    def __init__(
        self,
        name: str,
        positional: bool = False,
        categories: Sequence[str] = None,
        env: str = None,
        description: str = None,
        default: str = None,
        parser: Callable[[str], Any] = str,
    ):
        assert name
        assert not name.startswith('-')
        assert name.count('/') <= 1
        if '/' not in name:
            self.name = name
            self.neg = None
        else:
            self.name, self.neg = name.split('/', maxsplit=1)
            assert self.neg, f'Expected boolean format --yes/--no'
            assert parser is str, f'Parser is forced for booleans to x => bool(int(x))'
            parser = lambda s: bool(int(s))
        self.positional = positional
        self.env = env
        self.description = description
        self.default = default
        self.categories = categories
        self.parser = parser

    def parse(self, value: Optional[str] = None):
        if value == None:
            default = self.default
            if self.env:
                default = os.getenv(self.env, default)
            if default == None:
                return None
            value = str(default)
        return self.parser(value)

    def usage(self):
        default = f'[default: {self.default}]' if self.default else ''
        if self.env:
            value = os.getenv(self.env)
            value = '' if value is None else f':{value}'
            env = f'(env {self.env}{value}) '
        else:
            env = ''
        name = self.name
        if self.neg:
            name += f'/--{self.neg}'
        key = (f'--{name}' + ' ' * 30)[:15]
        description = '' if self.description is None else self.description
        return f'{key} {env}{default} {description}'

    def params(self):
        arg = f'--{self.name}'
        suffix = '' if self.positional else ' ...'
        if self.neg:
            arg += f'/--{self.neg}'
            suffix = ''
        if self.categories:
            suffix = ': ' + '/'.join(self.categories)
        if self.positional:
            return f'<{self.name}{suffix}>'
        else:
            return f'[{arg}{suffix}]'


class CommandDuck(abc.ABC):

    options: Sequence[Option]

    def __call__(self):
        script = sys.argv[0]
        out = self.call(script, *sys.argv[1:])
        if isinstance(out, str):
            print(out)
        else:
            pprint.pprint(out, compact=True)
        return

    @abc.abstractmethod
    def parse(self, prefix: str, *argv: str) -> \
        Tuple[Callable, Sequence, Dict]:
        ...

    @abc.abstractmethod
    def usage(self, prefix: str) -> str:
        ...

    def call(self, prefix: str, *argv: str):
        try:
            func, args, kwargs = self.parse(prefix, *argv)
        except AssertionError as e:
            red = lambda s: f'\033[0;31m{s}\033[00m'
            print(self.usage(prefix))
            print(red('Error: '), *map(red, e.args))
            sys.exit(1)
        return func(*args, **kwargs)

    def params(self):
        options = [opt.params() for opt in self.options]
        return ' '.join(options)


class Command(CommandDuck):

    def __init__(
        self,
        func: Callable,
        *options: Option,
        name: str = None,
        docs: str = None,
    ):
        self.func = func
        self.options = options
        self.name = func.__name__ if name is None else name
        self.docs = func.__doc__ if docs is None else docs

    def parse(self, prefix: str, *argv: str):
        options = {opt.name: opt for opt in self.options}
        positional = (opt for opt in self.options if opt.positional)
        neg_options = {opt.neg: opt for opt in self.options if opt.name}

        str_kwargs: Dict[str, Any] = {}
        it = iter(argv)
        for arg in it:
            if arg.startswith('--'):
                key = arg[2:]
                opt = options.get(key, neg_options.get(key))
                if opt and opt.neg:
                    value = '0' if key == opt.neg else '1'
                elif opt:
                    value = next(it, None)
                    assert value is not None, f'Expecting value after {arg}'
                elif arg == '--help':
                    return self.usage, [prefix], {}
                else:
                    assert False, f'Unknown option {arg}'
            else:
                value = arg
                opt = next(positional, None)
                assert opt, f'Unexpected positional value {value}'
            str_kwargs[opt.name] = value

        kwargs: Dict[str, Any] = {}
        for opt in options.values():
            value = opt.parse(str_kwargs.get(opt.name))
            assert value is not None, f'Required argument {opt.name} has no value'
            kwargs[opt.name] = value
        return self.func, [], kwargs

    def usage(self, prefix: str):
        options = '\n    '.join(opt.usage() for opt in self.options)
        docs = '' if self.docs is None else f'Docs:\n{self.docs}'
        return f'''
        Usage:
            {prefix} {self.params()}
            {prefix} --help
        
        Options:
            {options}
        {docs}
        '''.replace('\n        ', '\n').strip()


class CommandDict(Dict[str, Command], CommandDuck):

    def __init__(self, *commands: Command, **renamed_commands: Command):
        self.categories = []
        self.options = [
            Option('command', positional=True, categories=self.categories)
        ]
        self.add(*commands, **renamed_commands)

    def add(self, *commands: Command, **renamed_commands: Command):
        for command in commands:
            self[command.name] = command
            self.categories.append(command.name)
        for name, command in renamed_commands.items():
            self[name] = command
            self.categories.append(name)

    def parse(self, prefix: str, *argv: str):
        assert argv, f'Expected a command from: {list(self)}'
        first, *more = argv
        if first == '--help':
            return self.usage, [prefix], {}
        cmd = self.get(first)
        assert cmd, f'Command {first} not in {list(self)}\n'
        return cmd.call, [f'{prefix} {first}', *more], {}

    def usage(self, prefix: str):
        options = '\n'.join(
            [f'    {(k+" "*20)[:15]}{v.params()}' for k, v in self.items()])
        usages = '\n'.join([
            f'    {prefix} <command> [...command arguments]',
            f'    {prefix} <command> --help',
            f'    {prefix} --help',
        ])
        return f'Usage:\n{usages}\n\nOptions for <command>:\n{options}'



from typing import Dict, List, Tuple, TypeVar, Union

Data = Union[str, float, int, bool, None]
Record = Dict[str, Data]
Params = List[Data]
DataRow = List[Data]

K = TypeVar('K')
V = TypeVar('V')


def unzip(record: Dict[K, V]) -> Tuple[List[K], List[V]]:
    return map(list, zip(*record.items()))  # type:ignore
